@@ -27,6 +27,9 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
 #ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
@@ -70,6 +73,16 @@
 #define SOCKET int
 #endif
 
+
+#ifdef __MINGW32__
+#define _WINSOCKAPI_
+#include <wsockcompat.h>
+#include <winsock2.h>
+#undef SOCKLEN_T
+#define SOCKLEN_T unsigned int
+#endif
+
+
 #include <libxml/globals.h>
 #include <libxml/xmlerror.h>
 #include <libxml/xmlmemory.h>
@@ -82,10 +95,18 @@
  * A couple portability macros
  */
 #ifndef _WINSOCKAPI_
+#ifndef __BEOS__
 #define closesocket(s) close(s)
+#endif
 #define SOCKET int
 #endif
 
+#ifdef __BEOS__
+#ifndef PF_INET
+#define PF_INET AF_INET
+#endif
+#endif
+
 #ifndef SOCKLEN_T
 #define SOCKLEN_T unsigned int
 #endif
@@ -128,6 +149,8 @@ typedef struct xmlNanoHTTPCtxt {
     char *contentType;	/* the MIME type for the input */
     char *location;	/* the new URL in case of redirect */
     char *authHeader;	/* contents of {WWW,Proxy}-Authenticate header */
+    char *encoding;	/* encoding extracted from the contentType */
+    char *mimeType;	/* Mime-Type extracted from the contentType */
 } xmlNanoHTTPCtxt, *xmlNanoHTTPCtxtPtr;
 
 static int initialized = 0;
@@ -138,6 +161,18 @@ static unsigned int timeout = 60;/* the select() timeout in seconds */
 int xmlNanoHTTPFetchContent( void * ctx, char ** ptr, int * len );
 int xmlNanoHTTPContentLength( void * ctx );
 
+/**
+ * xmlHTTPErrMemory:
+ * @extra:  extra informations
+ *
+ * Handle an out of memory condition
+ */
+static void
+xmlHTTPErrMemory(const char *extra)
+{
+    __xmlSimpleError(XML_FROM_HTTP, XML_ERR_NO_MEMORY, NULL, NULL, extra);
+}
+
 /**
  * A portability function
  */
@@ -149,6 +184,20 @@ static int socket_errno(void) {
 #endif
 }
 
+#ifdef SUPPORT_IP6
+static
+int have_ipv6(void) {
+    int s;
+
+    s = socket (AF_INET6, SOCK_STREAM, 0);
+    if (s != -1) {
+	close (s);
+	return (1);
+    }
+    return (0);
+}
+#endif
+
 /**
  * xmlNanoHTTPInit:
  *
@@ -253,26 +302,64 @@ xmlNanoHTTPScanURL(xmlNanoHTTPCtxtPtr ctxt, const char *URL) {
 
     buf[indx] = 0;
     while (1) {
-        if (cur[0] == ':') {
+	if ((strchr (cur, '[') && !strchr (cur, ']')) ||
+		(!strchr (cur, '[') && strchr (cur, ']'))) {
+	    __xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, 
+	               "Syntax Error\n");
+	    return;
+	}
+
+	if (cur[0] == '[') {
+	    cur++;
+	    while (cur[0] != ']')
+		buf[indx++] = *cur++;
+    
+	    if (!strchr (buf, ':')) {
+		__xmlIOErr(XML_FROM_HTTP, XML_HTTP_USE_IP,
+			"Use [IPv6]/IPv4 format\n");
+		return;
+	    }
+
 	    buf[indx] = 0;
-	    ctxt->hostname = xmlMemStrdup(buf);
+	    ctxt->hostname = xmlMemStrdup (buf);
 	    indx = 0;
 	    cur += 1;
-	    while ((*cur >= '0') && (*cur <= '9')) {
-	        port *= 10;
-		port += *cur - '0';
+	    if (cur[0] == ':') {
 		cur++;
+		while (*cur >= '0' && *cur <= '9') {
+		    port *= 10;
+		    port += *cur - '0';
+		    cur++;
+		}
+
+		if (port != 0) ctxt->port = port;
+		while ((cur[0] != '/') && (*cur != 0))
+		    cur++;
 	    }
-	    if (port != 0) ctxt->port = port;
-	    while ((cur[0] != '/') && (*cur != 0)) 
-	        cur++;
-	    break;
-	}
-        if ((*cur == '/') || (*cur == 0)) {
-	    buf[indx] = 0;
-	    ctxt->hostname = xmlMemStrdup(buf);
-	    indx = 0;
 	    break;
+        }
+	else {
+	    if (cur[0] == ':') {
+		buf[indx] = 0;
+		ctxt->hostname = xmlMemStrdup (buf);
+		indx = 0;
+		cur += 1;
+		while ((*cur >= '0') && (*cur <= '9')) {
+		    port *= 10;
+		    port += *cur - '0';
+		    cur++;
+		}
+		if (port != 0) ctxt->port = port;
+		while ((cur[0] != '/') && (*cur != 0)) 
+		    cur++;
+		break;
+	    }
+	    if ((*cur == '/') || (*cur == 0)) {
+		buf[indx] = 0;
+		ctxt->hostname = xmlMemStrdup (buf);
+		indx = 0;
+		break;
+	    }
 	}
 	buf[indx++] = *cur++;
     }
@@ -335,26 +422,63 @@ xmlNanoHTTPScanProxy(const char *URL) {
 
     buf[indx] = 0;
     while (1) {
-        if (cur[0] == ':') {
+	if ((strchr (cur, '[') && !strchr (cur, ']')) ||
+		(!strchr (cur, '[') && strchr (cur, ']'))) {
+	    __xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, "Syntax Error\n");
+	    return;
+	}
+
+	if (cur[0] == '[') {
+	    cur++;
+	    while (cur[0] != ']')
+		buf[indx++] = *cur++;
+
+	    if (!strchr (buf, ':')) {
+		__xmlIOErr(XML_FROM_HTTP, XML_HTTP_USE_IP,
+			"Use [IPv6]/IPv4 format\n");
+		return;
+	    }
+
 	    buf[indx] = 0;
-	    proxy = xmlMemStrdup(buf);
+	    proxy = xmlMemStrdup (buf);
 	    indx = 0;
 	    cur += 1;
-	    while ((*cur >= '0') && (*cur <= '9')) {
-	        port *= 10;
-		port += *cur - '0';
-		cur++;
-	    }
-	    if (port != 0) proxyPort = port;
-	    while ((cur[0] != '/') && (*cur != 0)) 
+	    if (cur[0] == ':') {
 	        cur++;
+		while (*cur >= '0' && *cur <= '9') {
+		    port *= 10;
+		    port += *cur - '0';
+		    cur++;
+		}
+
+		if (port != 0) proxyPort = port;
+		while ((cur[0] != '/') && (*cur != 0))
+		    cur ++;
+	    }
 	    break;
 	}
-        if ((*cur == '/') || (*cur == 0)) {
-	    buf[indx] = 0;
-	    proxy = xmlMemStrdup(buf);
-	    indx = 0;
-	    break;
+	else {
+	    if (cur[0] == ':') {
+		buf[indx] = 0;
+		proxy = xmlMemStrdup (buf);
+		indx = 0;
+		cur += 1;
+		while ((*cur >= '0') && (*cur <= '9')) {
+		    port *= 10;
+		    port += *cur - '0';
+		    cur++;
+		}
+		if (port != 0) proxyPort = port;
+		while ((cur[0] != '/') && (*cur != 0)) 
+		    cur++;
+		break;
+	    }
+	    if ((*cur == '/') || (*cur == 0)) {
+		buf[indx] = 0;
+		proxy = xmlMemStrdup (buf);
+		indx = 0;
+		break;
+	    }
 	}
 	buf[indx++] = *cur++;
     }
@@ -374,7 +498,10 @@ xmlNanoHTTPNewCtxt(const char *URL) {
     xmlNanoHTTPCtxtPtr ret;
 
     ret = (xmlNanoHTTPCtxtPtr) xmlMalloc(sizeof(xmlNanoHTTPCtxt));
-    if (ret == NULL) return(NULL);
+    if (ret == NULL) {
+        xmlHTTPErrMemory("allocating context");
+        return(NULL);
+    }
 
     memset(ret, 0, sizeof(xmlNanoHTTPCtxt));
     ret->port = 80;
@@ -403,6 +530,8 @@ xmlNanoHTTPFreeCtxt(xmlNanoHTTPCtxtPtr ctxt) {
     if (ctxt->out != NULL) xmlFree(ctxt->out);
     if (ctxt->in != NULL) xmlFree(ctxt->in);
     if (ctxt->contentType != NULL) xmlFree(ctxt->contentType);
+    if (ctxt->encoding != NULL) xmlFree(ctxt->encoding);
+    if (ctxt->mimeType != NULL) xmlFree(ctxt->mimeType);
     if (ctxt->location != NULL) xmlFree(ctxt->location);
     if (ctxt->authHeader != NULL) xmlFree(ctxt->authHeader);
     ctxt->state = XML_NANO_HTTP_NONE;
@@ -435,10 +564,7 @@ xmlNanoHTTPSend(xmlNanoHTTPCtxtPtr ctxt, const char * xmt_ptr, int outlen) {
 	    	      ( socket_errno( ) != EAGAIN ) &&
 #endif
 		        ( socket_errno( ) != EWOULDBLOCK ) ) {
-	        xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPSend error:  %s",
-				strerror( socket_errno( ) ) );
-
+		__xmlIOErr(XML_FROM_HTTP, 0, "send failed\n");
 		if ( total_sent == 0 )
 		    total_sent = -1;
 		break;
@@ -486,9 +612,8 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 	if (ctxt->in == NULL) {
 	    ctxt->in = (char *) xmlMallocAtomic(65000 * sizeof(char));
 	    if (ctxt->in == NULL) {
+		xmlHTTPErrMemory("allocating input");
 	        ctxt->last = -1;
-		xmlGenericError( xmlGenericErrorContext, 
-			"xmlNanoHTTPRecv:  Error allocating input memory." );
 		return(-1);
 	    }
 	    ctxt->inlen = 65000;
@@ -512,10 +637,7 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 	    ctxt->inlen *= 2;
             ctxt->in = (char *) xmlRealloc(tmp_ptr, ctxt->inlen);
 	    if (ctxt->in == NULL) {
-	        xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPRecv:  %s %d bytes.",
-				"Failed to realloc input buffer to",
-				ctxt->inlen );
+		xmlHTTPErrMemory("allocating input buffer");
 		xmlFree( tmp_ptr );
 	        ctxt->last = -1;
 		return(-1);
@@ -546,9 +668,7 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 		    return ( 0 );
 
 		default:
-		    xmlGenericError( xmlGenericErrorContext,
-		    		"xmlNanoHTTPRecv:  recv( ) failure - %s",
-				strerror( socket_errno( ) ) );
+		    __xmlIOErr(XML_FROM_HTTP, 0, "recv failed\n");
 		    return(-1);
 	    }
 	}
@@ -619,7 +739,7 @@ xmlNanoHTTPReadLine(xmlNanoHTTPCtxtPtr ctxt) {
  * Try to extract useful informations from the server answer.
  * We currently parse and process:
  *  - The HTTP revision/ return code
- *  - The Content-Type
+ *  - The Content-Type, Mime-Type and charset used
  *  - The Location for redirect processing.
  *
  * Returns -1 in case of failure, the file descriptor number otherwise
@@ -663,16 +783,56 @@ xmlNanoHTTPScanAnswer(xmlNanoHTTPCtxtPtr ctxt, const char *line) {
 	if ((*cur != 0) && (*cur != ' ') && (*cur != '\t')) return;
 	ctxt->returnValue = ret;
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"Content-Type:", 13)) {
+        const xmlChar *charset, *last, *mime;
         cur += 13;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
 	if (ctxt->contentType != NULL)
 	    xmlFree(ctxt->contentType);
 	ctxt->contentType = xmlMemStrdup(cur);
+	mime = (const xmlChar *) cur;
+	last = mime;
+	while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	       (*last != ';') && (*last != ','))
+	    last++;
+	if (ctxt->mimeType != NULL)
+	    xmlFree(ctxt->mimeType);
+	ctxt->mimeType = (char *) xmlStrndup(mime, last - mime);
+	charset = xmlStrstr(BAD_CAST ctxt->contentType, BAD_CAST "charset=");
+	if (charset != NULL) {
+	    charset += 8;
+	    last = charset;
+	    while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	           (*last != ';') && (*last != ','))
+		last++;
+	    if (ctxt->encoding != NULL)
+	        xmlFree(ctxt->encoding);
+	    ctxt->encoding = (char *) xmlStrndup(charset, last - charset);
+	}
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"ContentType:", 12)) {
+        const xmlChar *charset, *last, *mime;
         cur += 12;
 	if (ctxt->contentType != NULL) return;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
 	ctxt->contentType = xmlMemStrdup(cur);
+	mime = (const xmlChar *) cur;
+	last = mime;
+	while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	       (*last != ';') && (*last != ','))
+	    last++;
+	if (ctxt->mimeType != NULL)
+	    xmlFree(ctxt->mimeType);
+	ctxt->mimeType = (char *) xmlStrndup(mime, last - mime);
+	charset = xmlStrstr(BAD_CAST ctxt->contentType, BAD_CAST "charset=");
+	if (charset != NULL) {
+	    charset += 8;
+	    last = charset;
+	    while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	           (*last != ';') && (*last != ','))
+		last++;
+	    if (ctxt->encoding != NULL)
+	        xmlFree(ctxt->encoding);
+	    ctxt->encoding = (char *) xmlStrndup(charset, last - charset);
+	}
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"Location:", 9)) {
         cur += 9;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
@@ -711,19 +871,28 @@ xmlNanoHTTPScanAnswer(xmlNanoHTTPCtxtPtr ctxt, const char *line) {
 static int
 xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 {
-    SOCKET s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
     fd_set wfd;
     struct timeval tv;
     int status;
+    int addrlen;
+    SOCKET s;
     
+#ifdef SUPPORT_IP6
+    if (addr->sa_family == AF_INET6) {
+	s = socket (PF_INET6, SOCK_STREAM, IPPROTO_TCP);
+	addrlen = sizeof (struct sockaddr_in6);
+    }
+    else
+#endif
+    {
+	s = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
+	addrlen = sizeof (struct sockaddr_in);
+    }
     if (s==-1) {
 #ifdef DEBUG_HTTP
 	perror("socket");
 #endif
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPConnectAttempt: %s - %s",
-			"socket creation failure",
-			strerror( socket_errno( ) ) );
+	__xmlIOErr(XML_FROM_HTTP, 0, "socket failed\n");
 	return(-1);
     }
     
@@ -754,26 +923,20 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 #ifdef DEBUG_HTTP
 	perror("nonblocking");
 #endif
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPConnectAttempt:  %s - %s",
-			"error setting non-blocking IO",
-			strerror( socket_errno( ) ) );
+	__xmlIOErr(XML_FROM_HTTP, 0, "error setting non-blocking IO\n");
 	closesocket(s);
 	return(-1);
     }
 #endif /* !VMS */
 #endif /* !_WINSOCKAPI_ */
 
-    if ((connect(s, addr, sizeof(*addr))==-1)) {
+    if (connect (s, addr, addrlen) == -1) {
 	switch (socket_errno()) {
 	    case EINPROGRESS:
 	    case EWOULDBLOCK:
 		break;
 	    default:
-		xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPConnectAttempt:  %s - %s",
-				"error connecting to HTTP server",
-				strerror( socket_errno( ) ) );
+		__xmlIOErr(XML_FROM_HTTP, 0, "error connecting to HTTP server");
 		closesocket(s);
 		return(-1);
 	}
@@ -789,17 +952,12 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
     {
 	case 0:
 	    /* Time out */
-	    xmlGenericError( xmlGenericErrorContext, 
-	    			"xmlNanoHTTPConnectAttempt: %s",
-				"Connect attempt timed out." );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Connect attempt timed out");
 	    closesocket(s);
 	    return(-1);
 	case -1:
 	    /* Ermm.. ?? */
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error connecting to host",
-				strerror( socket_errno( ) ) );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Connect failed");
 	    closesocket(s);
 	    return(-1);
     }
@@ -807,28 +965,22 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
     if ( FD_ISSET(s, &wfd) ) {
 	SOCKLEN_T len;
 	len = sizeof(status);
+#ifdef SO_ERROR
 	if (getsockopt(s, SOL_SOCKET, SO_ERROR, (char*)&status, &len) < 0 ) {
 	    /* Solaris error code */
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error retrieving pending socket errors",
-				strerror( socket_errno( ) ) );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "getsockopt failed\n");
 	    return (-1);
 	}
+#endif
 	if ( status ) {
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Error connecting to remote host");
 	    closesocket(s);
 	    errno = status;
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error connecting to remote host",
-				strerror( status ) );
 	    return (-1);
 	}
     } else {
 	/* pbm */
-	xmlGenericError( xmlGenericErrorContext,
-		"xmlNanoHTTPConnectAttempt:  %s\n",
-		"Select returned, but descriptor not set for connection.\n" );
+	__xmlIOErr(XML_FROM_HTTP, 0, "select failed\n");
 	closesocket(s);
 	return (-1);
     }
@@ -851,7 +1003,7 @@ static int
 xmlNanoHTTPConnectHost(const char *host, int port)
 {
     struct hostent *h;
-    struct sockaddr *addr;
+    struct sockaddr *addr = NULL;
     struct in_addr ia;
     struct sockaddr_in sockin;
 
@@ -862,13 +1014,60 @@ xmlNanoHTTPConnectHost(const char *host, int port)
     int i;
     int s;
 
-#if defined(SUPPORT_IP6) && defined(RES_USE_INET6)
-    if (!(_res.options & RES_INIT))
-        res_init();
-    _res.options |= RES_USE_INET6;
+    memset (&sockin, 0, sizeof(sockin));
+#ifdef SUPPORT_IP6
+    memset (&sockin6, 0, sizeof(sockin6));
+    if (have_ipv6 ())
+#if !defined(HAVE_GETADDRINFO) && defined(RES_USE_INET6)
+    {
+	if (!(_res.options & RES_INIT))
+	    res_init();
+	_res.options |= RES_USE_INET6;
+    }
+#elif defined(HAVE_GETADDRINFO)
+    {
+	int status;
+	struct addrinfo hints, *res, *result;
+
+	result = NULL;
+	memset (&hints, 0,sizeof(hints));
+	hints.ai_socktype = SOCK_STREAM;
+
+	status = getaddrinfo (host, NULL, &hints, &result);
+	if (status) {
+	    __xmlIOErr(XML_FROM_HTTP, 0, "getaddrinfo failed\n");
+	    return (-1);
+	}
+
+	for (res = result; res; res = res->ai_next) {
+	    if (res->ai_family == AF_INET || res->ai_family == AF_INET6) {
+		if (res->ai_family == AF_INET6) {
+		    memcpy (&sockin6, res->ai_addr, res->ai_addrlen);
+		    sockin6.sin6_port = htons (port);
+		    addr = (struct sockaddr *)&sockin6;
+		}
+		else {
+		    memcpy (&sockin, res->ai_addr, res->ai_addrlen);
+		    sockin.sin_port = htons (port);
+		    addr = (struct sockaddr *)&sockin;
+		}
+
+		s = xmlNanoHTTPConnectAttempt (addr);
+		if (s != -1) {
+		    freeaddrinfo (result);
+		    return (s);
+		}
+	    }
+	}
+	if (result)
+	    freeaddrinfo (result);
+	return (-1);
+    } else
 #endif
-    h = gethostbyname(host);
-    if (h == NULL) {
+#endif
+    {   
+	h = gethostbyname (host);
+	if (h == NULL) {
 
 /*
  * Okay, I got fed up by the non-portability of this error message
@@ -876,68 +1075,64 @@ xmlNanoHTTPConnectHost(const char *host, int port)
  * and one want to enable it, send me the defined(foobar) needed
  */
 #if defined(HAVE_NETDB_H) && defined(HOST_NOT_FOUND) && defined(linux)
-        const char *h_err_txt = "";
+	    const char *h_err_txt = "";
 
-        switch (h_errno) {
-            case HOST_NOT_FOUND:
-                h_err_txt = "Authoritive host not found";
-                break;
+	    switch (h_errno) {
+		case HOST_NOT_FOUND:
+		    h_err_txt = "Authoritive host not found";
+		    break;
 
-            case TRY_AGAIN:
-                h_err_txt =
-                    "Non-authoritive host not found or server failure.";
-                break;
+		case TRY_AGAIN:
+		    h_err_txt =
+			"Non-authoritive host not found or server failure.";
+		    break;
 
-            case NO_RECOVERY:
-                h_err_txt =
-                    "Non-recoverable errors:  FORMERR, REFUSED, or NOTIMP.";
-                break;
+		case NO_RECOVERY:
+		    h_err_txt =
+			"Non-recoverable errors:  FORMERR, REFUSED, or NOTIMP.";
+		    break;
 
-            case NO_ADDRESS:
-                h_err_txt =
-                    "Valid name, no data record of requested type.";
-                break;
+		case NO_ADDRESS:
+		    h_err_txt =
+			"Valid name, no data record of requested type.";
+		    break;
 
-            default:
-                h_err_txt = "No error text defined.";
-                break;
-        }
-        xmlGenericError(xmlGenericErrorContext,
-                        "xmlNanoHTTPConnectHost:  %s '%s' - %s",
-                        "Failed to resolve host", host, h_err_txt);
+		default:
+		    h_err_txt = "No error text defined.";
+		    break;
+	    }
+	    __xmlIOErr(XML_FROM_HTTP, 0, h_err_txt);
 #else
-        xmlGenericError(xmlGenericErrorContext,
-                        "xmlNanoHTTPConnectHost:  %s '%s'",
-                        "Failed to resolve host", host);
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Failed to resolve host");
 #endif
-        return (-1);
-    }
+	    return (-1);
+	}
 
-    for (i = 0; h->h_addr_list[i]; i++) {
-        if (h->h_addrtype == AF_INET) {
-            /* A records (IPv4) */
-            memcpy(&ia, h->h_addr_list[i], h->h_length);
-            sockin.sin_family = h->h_addrtype;
-            sockin.sin_addr = ia;
-            sockin.sin_port = htons(port);
-            addr = (struct sockaddr *) &sockin;
+	for (i = 0; h->h_addr_list[i]; i++) {
+	    if (h->h_addrtype == AF_INET) {
+		/* A records (IPv4) */
+		memcpy (&ia, h->h_addr_list[i], h->h_length);
+		sockin.sin_family = h->h_addrtype;
+		sockin.sin_addr = ia;
+		sockin.sin_port = htons (port);
+		addr = (struct sockaddr *) &sockin;
 #ifdef SUPPORT_IP6
-        } else if (h->h_addrtype == AF_INET6) {
-            /* AAAA records (IPv6) */
-            memcpy(&ia6, h->h_addr_list[i], h->h_length);
-            sockin6.sin_family = h->h_addrtype;
-            sockin6.sin_addr = ia6;
-            sockin6.sin_port = htons(port);
-            addr = (struct sockaddr *) &sockin6;
+	    } else if (have_ipv6 () && (h->h_addrtype == AF_INET6)) {
+		/* AAAA records (IPv6) */
+		memcpy (&ia6, h->h_addr_list[i], h->h_length);
+		sockin6.sin6_family = h->h_addrtype;
+		sockin6.sin6_addr = ia6;
+		sockin6.sin6_port = htons (port);
+		addr = (struct sockaddr *) &sockin6;
 #endif
-        } else
-            break;              /* for */
+	    } else
+		break;              /* for */
 
-        s = xmlNanoHTTPConnectAttempt(addr);
-        if (s != -1)
-            return (s);
+	    s = xmlNanoHTTPConnectAttempt (addr);
+	    if (s != -1)
+		return (s);
+	}
     }
-
 #ifdef DEBUG_HTTP
     xmlGenericError(xmlGenericErrorContext,
                     "xmlNanoHTTPConnectHost:  unable to connect to '%s'.\n",
@@ -1059,9 +1254,11 @@ xmlNanoHTTPMethodRedir(const char *URL, const char *method, const char *input,
     char *bp, *p;
     int blen, ret;
     int head;
-    int xmt_bytes;
     int nbRedirects = 0;
     char *redirURL = NULL;
+#ifdef DEBUG_HTTP
+    int xmt_bytes;
+#endif
     
     if (URL == NULL) return(NULL);
     if (method == NULL) method = "GET";
@@ -1072,30 +1269,22 @@ retry:
 	ctxt = xmlNanoHTTPNewCtxt(URL);
     else {
 	ctxt = xmlNanoHTTPNewCtxt(redirURL);
+	ctxt->location = xmlMemStrdup(redirURL);
     }
 
     if ( ctxt == NULL ) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s %s.",
-			"Unable to allocate HTTP context to URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
 	return ( NULL );
     }
 
     if ((ctxt->protocol == NULL) || (strcmp(ctxt->protocol, "http"))) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s - %s.",
-			"Not a valid HTTP URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
+	__xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, "Not a valid HTTP URI");
         xmlNanoHTTPFreeCtxt(ctxt);
 	if (redirURL != NULL) xmlFree(redirURL);
         return(NULL);
     }
     if (ctxt->hostname == NULL) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s - %s",
-			"Failed to identify host in URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
+	__xmlIOErr(XML_FROM_HTTP, XML_HTTP_UNKNOWN_HOST,
+	           "Failed to identify host in URI");
         xmlNanoHTTPFreeCtxt(ctxt);
 	if (redirURL != NULL) xmlFree(redirURL);
         return(NULL);
@@ -1128,9 +1317,7 @@ retry:
     bp = xmlMallocAtomic(blen);
     if ( bp == NULL ) {
         xmlNanoHTTPFreeCtxt( ctxt );
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s",
-			"Error allocating HTTP header buffer." );
+	xmlHTTPErrMemory("allocating header buffer");
 	return ( NULL );
     }
 
@@ -1173,26 +1360,30 @@ retry:
     ctxt->outptr = ctxt->out = bp;
     ctxt->state = XML_NANO_HTTP_WRITE;
     blen = strlen( ctxt->out );
-    xmt_bytes = xmlNanoHTTPSend(ctxt, ctxt->out, blen );
 #ifdef DEBUG_HTTP
+    xmt_bytes = xmlNanoHTTPSend(ctxt, ctxt->out, blen );
     if ( xmt_bytes != blen )
         xmlGenericError( xmlGenericErrorContext,
 			"xmlNanoHTTPMethodRedir:  Only %d of %d %s %s\n",
 			xmt_bytes, blen,
 			"bytes of HTTP headers sent to host",
 			ctxt->hostname );
+#else
+    xmlNanoHTTPSend(ctxt, ctxt->out, blen );
 #endif
 
     if ( input != NULL ) {
+#ifdef DEBUG_HTTP
         xmt_bytes = xmlNanoHTTPSend( ctxt, input, ilen );
 
-#ifdef DEBUG_HTTP
 	if ( xmt_bytes != ilen )
 	    xmlGenericError( xmlGenericErrorContext,
 	    		"xmlNanoHTTPMethodRedir:  Only %d of %d %s %s\n",
 			xmt_bytes, ilen,
 			"bytes of HTTP content sent to host",
 			ctxt->hostname );
+#else
+	xmlNanoHTTPSend( ctxt, input, ilen );
 #endif
     }
 
@@ -1339,6 +1530,7 @@ xmlNanoHTTPFetch(const char *URL, const char *filename, char **contentType) {
     return(0);
 }
 
+#ifdef LIBXML_OUTPUT_ENABLED
 /**
  * xmlNanoHTTPSave:
  * @ctxt:  the HTTP context
@@ -1375,6 +1567,7 @@ xmlNanoHTTPSave(void *ctxt, const char *filename) {
     xmlNanoHTTPClose(ctxt);
     return(0);
 }
+#endif /* LIBXML_OUTPUT_ENABLED */
 
 /**
  * xmlNanoHTTPReturnCode:
@@ -1428,6 +1621,51 @@ xmlNanoHTTPContentLength( void * ctx ) {
     return ( ( ctxt == NULL ) ? -1 : ctxt->ContentLength );
 }
 
+/**
+ * xmlNanoHTTPRedir:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified redirection URL if available from the HTTP header.
+ *
+ * Return the specified redirection URL or NULL if not redirected.
+ */
+const char *
+xmlNanoHTTPRedir( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->location );
+}
+
+/**
+ * xmlNanoHTTPEncoding:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified encoding if specified in the HTTP headers.
+ *
+ * Return the specified encoding or NULL if not available
+ */
+const char *
+xmlNanoHTTPEncoding( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->encoding );
+}
+
+/**
+ * xmlNanoHTTPMimeType:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified Mime-Type if specified in the HTTP headers.
+ *
+ * Return the specified Mime-Type or NULL if not available
+ */
+const char *
+xmlNanoHTTPMimeType( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->mimeType );
+}
+
 /**
  * xmlNanoHTTPFetchContent:
  * @ctx:  the HTTP context
