@@ -10,26 +10,30 @@
 
 #include <string.h>
 #include <stdarg.h>
-
 #include <assert.h>
 
 #if defined (_WIN32) && !defined(__CYGWIN__)
-#ifdef _MSC_VER
+#if defined (_MSC_VER) || defined(__BORLANDC__)
 #include <winsock2.h>
 #pragma comment(lib, "ws2_32.lib")
 #define gettimeofday(p1,p2)
-#include <time.h>
-#else /* _MSC_VER */
-#include <sys/time.h>
 #endif /* _MSC_VER */
-#else /* _WIN32 */
+#endif /* _WIN32 */
+
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
 #ifdef HAVE_TIME_H
 #include <time.h>
 #endif
-#endif /* _WIN32 */
+
+#ifdef __MINGW32__
+#define _WINSOCKAPI_
+#include <wsockcompat.h>
+#include <winsock2.h>
+#undef SOCKLEN_T
+#define SOCKLEN_T unsigned int
+#endif
 
 #ifdef HAVE_SYS_TIMEB_H
 #include <sys/timeb.h>
@@ -79,9 +83,6 @@
 #ifdef LIBXML_CATALOG_ENABLED
 #include <libxml/catalog.h>
 #endif
-#ifdef LIBXML_DOCB_ENABLED
-#include <libxml/DOCBparser.h>
-#endif
 #include <libxml/globals.h>
 #include <libxml/xmlreader.h>
 #ifdef LIBXML_SCHEMAS_ENABLED
@@ -98,14 +99,25 @@ static int shell = 0;
 static int debugent = 0;
 #endif
 static int debug = 0;
+#ifdef LIBXML_TREE_ENABLED
 static int copy = 0;
+#endif /* LIBXML_TREE_ENABLED */
 static int recovery = 0;
 static int noent = 0;
+static int noblanks = 0;
 static int noout = 0;
 static int nowrap = 0;
+#ifdef LIBXML_OUTPUT_ENABLED
+static int format = 0;
+static const char *output = NULL;
+static int compress = 0;
+#endif /* LIBXML_OUTPUT_ENABLED */
+#ifdef LIBXML_VALID_ENABLED
 static int valid = 0;
 static int postvalid = 0;
 static char * dtdvalid = NULL;
+static char * dtdvalidfpi = NULL;
+#endif
 #ifdef LIBXML_SCHEMAS_ENABLED
 static char * relaxng = NULL;
 static xmlRelaxNGPtr relaxngschemas = NULL;
@@ -114,18 +126,16 @@ static xmlSchemaPtr wxschemas = NULL;
 #endif
 static int repeat = 0;
 static int insert = 0;
-static int compress = 0;
-#ifdef LIBXML_DOCB_ENABLED
-static int sgml = 0;
-#endif
+#ifdef  LIBXML_HTML_ENABLED
 static int html = 0;
+#endif
 static int htmlout = 0;
+#ifdef LIBXML_PUSH_ENABLED
 static int push = 0;
+#endif /* LIBXML_PUSH_ENABLED */
 #ifdef HAVE_SYS_MMAN_H
 static int memory = 0;
 #endif
-static int noblanks = 0;
-static int format = 0;
 static int testIO = 0;
 static char *encoding = NULL;
 #ifdef LIBXML_XINCLUDE_ENABLED
@@ -141,10 +151,14 @@ static int dropdtd = 0;
 static int catalogs = 0;
 static int nocatalogs = 0;
 #endif
+#ifdef LIBXML_READER_ENABLED
 static int stream = 0;
+#endif /* LIBXML_READER_ENABLED */
 static int chkregister = 0;
-static const char *output = NULL;
-
+#ifdef LIBXML_SAX1_ENABLED
+static int sax1 = 0;
+#endif /* LIBXML_SAX1_ENABLED */
+static int options = 0;
 
 /*
  * Internal timing routines to remove the necessity to have unix-specific
@@ -382,14 +396,12 @@ xmlHTMLError(void *ctx, const char *msg, ...)
 {
     xmlParserCtxtPtr ctxt = (xmlParserCtxtPtr) ctx;
     xmlParserInputPtr input;
-    xmlParserInputPtr cur = NULL;
     va_list args;
     int len;
 
     buffer[0] = 0;
     input = ctxt->input;
     if ((input != NULL) && (input->filename == NULL) && (ctxt->inputNr > 1)) {
-	cur = input;
         input = ctxt->inputTab[ctxt->inputNr - 2];
     }
         
@@ -421,14 +433,12 @@ xmlHTMLWarning(void *ctx, const char *msg, ...)
 {
     xmlParserCtxtPtr ctxt = (xmlParserCtxtPtr) ctx;
     xmlParserInputPtr input;
-    xmlParserInputPtr cur = NULL;
     va_list args;
     int len;
 
     buffer[0] = 0;
     input = ctxt->input;
     if ((input != NULL) && (input->filename == NULL) && (ctxt->inputNr > 1)) {
-	cur = input;
         input = ctxt->inputTab[ctxt->inputNr - 2];
     }
         
@@ -582,22 +592,20 @@ static void myClose(FILE *f) {
   }
 }
 
+#ifdef LIBXML_READER_ENABLED
 /************************************************************************
  * 									*
  * 			Stream Test processing				*
  * 									*
  ************************************************************************/
-static int count = 0;
-static int elem, attrs;
-
 static void processNode(xmlTextReaderPtr reader) {
-    xmlChar *name, *value;
+    const xmlChar *name, *value;
 
-    name = xmlTextReaderName(reader);
+    name = xmlTextReaderConstName(reader);
     if (name == NULL)
 	name = xmlStrdup(BAD_CAST "--");
 
-    value = xmlTextReaderValue(reader);
+    value = xmlTextReaderConstValue(reader);
 
     printf("%d %d %s %d %d", 
 	    xmlTextReaderDepth(reader),
@@ -605,42 +613,58 @@ static void processNode(xmlTextReaderPtr reader) {
 	    name,
 	    xmlTextReaderIsEmptyElement(reader),
 	    xmlTextReaderHasValue(reader));
-    xmlFree(name);
     if (value == NULL)
 	printf("\n");
     else {
 	printf(" %s\n", value);
-	xmlFree(value);
     }
 }
 
 static void streamFile(char *filename) {
     xmlTextReaderPtr reader;
     int ret;
+#ifdef HAVE_SYS_MMAN_H
+    int fd = -1;
+    struct stat info;
+    const char *base = NULL;
+    xmlParserInputBufferPtr input = NULL;
 
-    if (count) {
-	elem = 0;
-	attrs = 0;
-    }
+    if (memory) {
+	if (stat(filename, &info) < 0) 
+	    return;
+	if ((fd = open(filename, O_RDONLY)) < 0)
+	    return;
+	base = mmap(NULL, info.st_size, PROT_READ, MAP_SHARED, fd, 0) ;
+	if (base == (void *) MAP_FAILED)
+	    return;
+
+	input = xmlParserInputBufferCreateStatic((char *) base, info.st_size,
+	                                      XML_CHAR_ENCODING_NONE);
+	reader = xmlNewTextReader(input, filename);
+    } else
+#endif
+	reader = xmlNewTextReaderFilename(filename);
 
-    reader = xmlNewTextReaderFilename(filename);
     if (reader != NULL) {
+#ifdef LIBXML_VALID_ENABLED
 	if (valid)
 	    xmlTextReaderSetParserProp(reader, XML_PARSER_VALIDATE, 1);
 	else
+#endif /* LIBXML_VALID_ENABLED */
 	    xmlTextReaderSetParserProp(reader, XML_PARSER_LOADDTD, 1);
 #ifdef LIBXML_SCHEMAS_ENABLED
 	if (relaxng != NULL) {
-	    if (timing) {
+	    if ((timing) && (!repeat)) {
 		startTimer();
 	    }
 	    ret = xmlTextReaderRelaxNGValidate(reader, relaxng);
 	    if (ret < 0) {
 		xmlGenericError(xmlGenericErrorContext,
 			"Relax-NG schema %s failed to compile\n", relaxng);
+		progresult = 5;
 		relaxng = NULL;
 	    }
-	    if (timing) {
+	    if ((timing) && (!repeat)) {
 		endTimer("Compiling the schemas");
 	    }
 	}
@@ -649,7 +673,7 @@ static void streamFile(char *filename) {
 	/*
 	 * Process all nodes in sequence
 	 */
-	if (timing) {
+	if ((timing) && (!repeat)) {
 	    startTimer();
 	}
 	ret = xmlTextReaderRead(reader);
@@ -658,17 +682,20 @@ static void streamFile(char *filename) {
 		processNode(reader);
 	    ret = xmlTextReaderRead(reader);
 	}
-	if (timing) {
+	if ((timing) && (!repeat)) {
 #ifdef LIBXML_SCHEMAS_ENABLED
 	    if ((valid) || (relaxng != NULL))
 #else
+#ifdef LIBXML_VALID_ENABLED
 	    if (valid)
-#endif
 		endTimer("Parsing and validating");
 	    else
+#endif /* LIBXML_VALID_ENABLED */
+#endif
 		endTimer("Parsing");
 	}
 
+#ifdef LIBXML_VALID_ENABLED
 	if (valid) {
 	    if (xmlTextReaderIsValid(reader) != 1) {
 		xmlGenericError(xmlGenericErrorContext,
@@ -676,6 +703,7 @@ static void streamFile(char *filename) {
 		progresult = 3;
 	    }
 	}
+#endif /* LIBXML_VALID_ENABLED */
 #ifdef LIBXML_SCHEMAS_ENABLED
 	if (relaxng != NULL) {
 	    if (xmlTextReaderIsValid(reader) != 1) {
@@ -698,20 +726,32 @@ static void streamFile(char *filename) {
 	fprintf(stderr, "Unable to open %s\n", filename);
 	progresult = 1;
     }
+#ifdef HAVE_SYS_MMAN_H
+    if (memory) {
+        xmlFreeParserInputBuffer(input);
+	munmap((char *) base, info.st_size);
+	close(fd);
+    }
+#endif
 }
+#endif /* LIBXML_READER_ENABLED */
 
 /************************************************************************
  * 									*
  * 			Tree Test processing				*
  * 									*
  ************************************************************************/
-static void parseAndPrintFile(char *filename) {
-    xmlDocPtr doc = NULL, tmp;
+static void parseAndPrintFile(char *filename, xmlParserCtxtPtr rectxt) {
+    xmlDocPtr doc = NULL;
+#ifdef LIBXML_TREE_ENABLED
+    xmlDocPtr tmp;
+#endif /* LIBXML_TREE_ENABLED */
 
     if ((timing) && (!repeat))
 	startTimer();
     
 
+#ifdef LIBXML_TREE_ENABLED
     if (filename == NULL) {
 	if (generate) {
 	    xmlNodePtr n;
@@ -722,44 +762,41 @@ static void parseAndPrintFile(char *filename) {
 	    xmlDocSetRootElement(doc, n);
 	}
     }
-#ifdef LIBXML_DOCB_ENABLED
-    /*
-     * build an SGML tree from a string;
-     */
-    else if ((sgml) && (push)) {
-	FILE *f;
-
-	f = fopen(filename, "r");
-	if (f != NULL) {
-	    int res, size = 3;
-	    char chars[4096];
-	    docbParserCtxtPtr ctxt;
-
-	    /* if (repeat) */
-		size = 4096;
-	    res = fread(chars, 1, 4, f);
-	    if (res > 0) {
-		ctxt = docbCreatePushParserCtxt(NULL, NULL,
-			    chars, res, filename, 0);
-		while ((res = fread(chars, 1, size, f)) > 0) {
-		    docbParseChunk(ctxt, chars, res, 0);
-		}
-		docbParseChunk(ctxt, chars, 0, 1);
-		doc = ctxt->myDoc;
-		docbFreeParserCtxt(ctxt);
-	    }
-	    fclose(f);
-	}
-    } else if (sgml) {	
-	doc = docbParseFile(filename, NULL);
-    }
-#endif
+#endif /* LIBXML_TREE_ENABLED */
 #ifdef LIBXML_HTML_ENABLED
+#ifdef LIBXML_PUSH_ENABLED
+    else if ((html) && (push)) {
+        FILE *f;
+
+        f = fopen(filename, "r");
+        if (f != NULL) {
+            int res, size = 3;
+            char chars[4096];
+            htmlParserCtxtPtr ctxt;
+
+            /* if (repeat) */
+                size = 4096;
+            res = fread(chars, 1, 4, f);
+            if (res > 0) {
+                ctxt = htmlCreatePushParserCtxt(NULL, NULL,
+                            chars, res, filename, XML_CHAR_ENCODING_NONE);
+                while ((res = fread(chars, 1, size, f)) > 0) {
+                    htmlParseChunk(ctxt, chars, res, 0);
+                }
+                htmlParseChunk(ctxt, chars, 0, 1);
+                doc = ctxt->myDoc;
+                htmlFreeParserCtxt(ctxt);
+            }
+            fclose(f);
+        }
+    }
+#endif /* LIBXML_PUSH_ENABLED */
     else if (html) {
-	doc = htmlParseFile(filename, NULL);
+	doc = htmlReadFile(filename, NULL, options);
     }
 #endif /* LIBXML_HTML_ENABLED */
     else {
+#ifdef LIBXML_PUSH_ENABLED
 	/*
 	 * build an XML tree from a string;
 	 */
@@ -796,74 +833,47 @@ static void parseAndPrintFile(char *filename) {
 		    }
 	        }
 	    }
-	} else if (testIO) {
-	    int ret;
-	    FILE *f;
-
-	    /* '-' Usually means stdin -<sven@zen.org> */
+	} else
+#endif /* LIBXML_PUSH_ENABLED */
+        if (testIO) {
 	    if ((filename[0] == '-') && (filename[1] == 0)) {
-	      f = stdin;
+	        doc = xmlReadFd(0, NULL, NULL, options);
 	    } else {
-	      f = fopen(filename, "r");
-	    }
-	    if (f != NULL) {
-                xmlParserCtxtPtr ctxt;
+	        FILE *f;
 
-		ctxt = xmlCreateIOParserCtxt(NULL, NULL,
-			    (xmlInputReadCallback) myRead,
-			    (xmlInputCloseCallback) myClose,
-			    f, XML_CHAR_ENCODING_NONE);
-		xmlParseDocument(ctxt);
-
-		ret = ctxt->wellFormed;
-		doc = ctxt->myDoc;
-		xmlFreeParserCtxt(ctxt);
-		if (!ret) {
-		    xmlFreeDoc(doc);
+		f = fopen(filename, "r");
+		if (f != NULL) {
+		    if (rectxt == NULL)
+			doc = xmlReadIO((xmlInputReadCallback) myRead,
+					(xmlInputCloseCallback) myClose, f,
+					filename, NULL, options);
+		    else
+			doc = xmlCtxtReadIO(rectxt,
+			                (xmlInputReadCallback) myRead,
+					(xmlInputCloseCallback) myClose, f,
+					filename, NULL, options);
+		} else
 		    doc = NULL;
-		}
 	    }
-	} else if (recovery) {
-	    doc = xmlRecoverFile(filename);
 	} else if (htmlout) {
-	    int ret;
 	    xmlParserCtxtPtr ctxt;
-	    xmlSAXHandler silent, *old;
-
-	    ctxt = xmlCreateFileParserCtxt(filename);
 
+	    if (rectxt == NULL)
+		ctxt = xmlNewParserCtxt();
+	    else
+	        ctxt = rectxt;
 	    if (ctxt == NULL) {	      
-	      /* If xmlCreateFileParserCtxt() return NULL something
-		 strange happened so we don't want to do anything.  Do
-		 we want to print an error message here?
-		 <sven@zen.org> */
-	      doc = NULL;
+	        doc = NULL;
 	    } else {
-	      memcpy(&silent, ctxt->sax, sizeof(silent));
-	      old = ctxt->sax;
-	      silent.error = xmlHTMLError;
-	      if (xmlGetWarningsDefaultValue)
-		silent.warning = xmlHTMLWarning;
-	      else 
-		silent.warning = NULL;
-	      silent.fatalError = xmlHTMLError;
-	      ctxt->sax = &silent;
-	      ctxt->vctxt.error = xmlHTMLValidityError;
-	      if (xmlGetWarningsDefaultValue)
-		ctxt->vctxt.warning = xmlHTMLValidityWarning;
-	      else 
-		ctxt->vctxt.warning = NULL;
+	        ctxt->sax->error = xmlHTMLError;
+	        ctxt->sax->warning = xmlHTMLWarning;
+	        ctxt->vctxt.error = xmlHTMLValidityError;
+	        ctxt->vctxt.warning = xmlHTMLValidityWarning;
 
-	      xmlParseDocument(ctxt);
+		doc = xmlCtxtReadFile(ctxt, filename, NULL, options);
 
-	      ret = ctxt->wellFormed;
-	      doc = ctxt->myDoc;
-	      ctxt->sax = old;
-	      xmlFreeParserCtxt(ctxt);
-	      if (!ret) {
-		xmlFreeDoc(doc);
-		doc = NULL;
-	      }
+		if (rectxt == NULL)
+		    xmlFreeParserCtxt(ctxt);
 	    }
 #ifdef HAVE_SYS_MMAN_H
 	} else if (memory) {
@@ -878,31 +888,39 @@ static void parseAndPrintFile(char *filename) {
 	    if (base == (void *) MAP_FAILED)
 	        return;
 
-	    doc = xmlParseMemory((char *) base, info.st_size);
+	    if (rectxt == NULL)
+		doc = xmlReadMemory((char *) base, info.st_size,
+		                    filename, NULL, options);
+	    else
+		doc = xmlCtxtReadMemory(rectxt, (char *) base, info.st_size,
+			                filename, NULL, options);
+	        
 	    munmap((char *) base, info.st_size);
 #endif
+#ifdef LIBXML_VALID_ENABLED
 	} else if (valid) {
-	    int ret;
-	    xmlParserCtxtPtr ctxt;
-
-	    ctxt = xmlCreateFileParserCtxt(filename);
+	    xmlParserCtxtPtr ctxt = NULL;
 
+	    if (rectxt == NULL)
+		ctxt = xmlNewParserCtxt();
+	    else
+	        ctxt = rectxt;
 	    if (ctxt == NULL) {	      
-	      doc = NULL;
+	        doc = NULL;
 	    } else {
-	      xmlParseDocument(ctxt);
-	      if (ctxt->valid == 0)
-		  progresult = 4;
-	      ret = ctxt->wellFormed;
-	      doc = ctxt->myDoc;
-	      xmlFreeParserCtxt(ctxt);
-	      if (!ret) {
-		xmlFreeDoc(doc);
-		doc = NULL;
-	      }
+		doc = xmlCtxtReadFile(ctxt, filename, NULL, options);
+
+		if (ctxt->valid == 0)
+		    progresult = 4;
+		if (rectxt == NULL)
+		    xmlFreeParserCtxt(ctxt);
 	    }
+#endif /* LIBXML_VALID_ENABLED */
 	} else {
-	    doc = xmlParseFile(filename);
+	    if (rectxt != NULL)
+	        doc = xmlCtxtReadFile(rectxt, filename, NULL, options);
+	    else
+		doc = xmlReadFile(filename, NULL, options);
 	}
     }
 
@@ -951,6 +969,7 @@ static void parseAndPrintFile(char *filename) {
         xmlShell(doc, filename, xmlShellReadline, stdout);
 #endif
 
+#ifdef LIBXML_TREE_ENABLED
     /*
      * test intermediate copy if needed.
      */
@@ -959,7 +978,9 @@ static void parseAndPrintFile(char *filename) {
 	doc = xmlCopyDoc(doc, 1);
 	xmlFreeDoc(tmp);
     }
+#endif /* LIBXML_TREE_ENABLED */
 
+#ifdef LIBXML_VALID_ENABLED
     if ((insert) && (!html)) {
         const xmlChar* list[256];
 	int nb, i;
@@ -983,7 +1004,10 @@ static void parseAndPrintFile(char *filename) {
 		}
 	    }
 	}    
-    }else if (noout == 0) {
+    }else
+#endif /* LIBXML_VALID_ENABLED */
+#ifdef LIBXML_OUTPUT_ENABLED
+    if (noout == 0) {
 	/*
 	 * print it.
 	 */
@@ -1039,10 +1063,16 @@ static void parseAndPrintFile(char *filename) {
 		else {
 		    out = fopen(output,"wb");
 		}
-		xmlDocDump(out, doc);
+		if (out != NULL) {
+		    if (xmlDocDump(out, doc) < 0)
+		        progresult = 6;
 
-		if (output)
-		    fclose(out);
+		    if (output != NULL)
+			fclose(out);
+		} else {
+		    fprintf(stderr, "failed to open %s\n", output);
+		    progresult = 6;
+		}
 	    }
 	    if ((timing) && (!repeat)) {
 		endTimer("Saving");
@@ -1055,30 +1085,44 @@ static void parseAndPrintFile(char *filename) {
 	    else {
 		out = fopen(output,"wb");
 	    }
-	    xmlDebugDumpDocument(out, doc);
+	    if (out != NULL) {
+		xmlDebugDumpDocument(out, doc);
 
-	    if (output)
-		fclose(out);
+		if (output != NULL)
+		    fclose(out);
+	    } else {
+		fprintf(stderr, "failed to open %s\n", output);
+		progresult = 6;
+	    }
 	}
 #endif
     }
+#endif /* LIBXML_OUTPUT_ENABLED */
 
+#ifdef LIBXML_VALID_ENABLED
     /*
      * A posteriori validation test
      */
-    if (dtdvalid != NULL) {
+    if ((dtdvalid != NULL) || (dtdvalidfpi != NULL)) {
 	xmlDtdPtr dtd;
 
 	if ((timing) && (!repeat)) {
 	    startTimer();
 	}
-	dtd = xmlParseDTD(NULL, (const xmlChar *)dtdvalid); 
+	if (dtdvalid != NULL)
+	    dtd = xmlParseDTD(NULL, (const xmlChar *)dtdvalid); 
+	else
+	    dtd = xmlParseDTD((const xmlChar *)dtdvalidfpi, NULL); 
 	if ((timing) && (!repeat)) {
 	    endTimer("Parsing DTD");
 	}
 	if (dtd == NULL) {
-	    xmlGenericError(xmlGenericErrorContext,
-		    "Could not parse DTD %s\n", dtdvalid);
+	    if (dtdvalid != NULL)
+		xmlGenericError(xmlGenericErrorContext,
+			"Could not parse DTD %s\n", dtdvalid);
+	    else
+		xmlGenericError(xmlGenericErrorContext,
+			"Could not parse DTD %s\n", dtdvalidfpi);
 	    progresult = 2;
 	} else {
 	    xmlValidCtxtPtr cvp;
@@ -1096,9 +1140,14 @@ static void parseAndPrintFile(char *filename) {
 		startTimer();
 	    }
 	    if (!xmlValidateDtd(cvp, doc, dtd)) {
-		xmlGenericError(xmlGenericErrorContext,
-			"Document %s does not validate against %s\n",
-			filename, dtdvalid);
+		if (dtdvalid != NULL)
+		    xmlGenericError(xmlGenericErrorContext,
+			    "Document %s does not validate against %s\n",
+			    filename, dtdvalid);
+		else
+		    xmlGenericError(xmlGenericErrorContext,
+			    "Document %s does not validate against %s\n",
+			    filename, dtdvalidfpi);
 		progresult = 3;
 	    }
 	    if ((timing) && (!repeat)) {
@@ -1131,8 +1180,10 @@ static void parseAndPrintFile(char *filename) {
 	    endTimer("Validating");
 	}
 	xmlFreeValidCtxt(cvp);
+    }
+#endif /* LIBXML_VALID_ENABLED */
 #ifdef LIBXML_SCHEMAS_ENABLED
-    } else if (relaxngschemas != NULL) {
+    if (relaxngschemas != NULL) {
 	xmlRelaxNGValidCtxtPtr ctxt;
 	int ret;
 
@@ -1184,8 +1235,8 @@ static void parseAndPrintFile(char *filename) {
 	if ((timing) && (!repeat)) {
 	    endTimer("Validating");
 	}
-#endif
     }
+#endif
 
 #ifdef LIBXML_DEBUG_ENABLED
     if ((debugent) && (!html))
@@ -1213,6 +1264,9 @@ static void parseAndPrintFile(char *filename) {
 static void showVersion(const char *name) {
     fprintf(stderr, "%s: using libxml version %s\n", name, xmlParserVersion);
     fprintf(stderr, "   compiled with: ");
+#ifdef LIBXML_VALID_ENABLED
+    fprintf(stderr, "DTDValid ");
+#endif
 #ifdef LIBXML_FTP_ENABLED
     fprintf(stderr, "FTP ");
 #endif
@@ -1228,9 +1282,6 @@ static void showVersion(const char *name) {
 #ifdef LIBXML_CATALOG_ENABLED
     fprintf(stderr, "Catalog ");
 #endif
-#ifdef LIBXML_DOCB_ENABLED
-    fprintf(stderr, "DocBook ");
-#endif
 #ifdef LIBXML_XPATH_ENABLED
     fprintf(stderr, "XPath ");
 #endif
@@ -1263,47 +1314,64 @@ static void showVersion(const char *name) {
 
 static void usage(const char *name) {
     printf("Usage : %s [options] XMLfiles ...\n", name);
+#ifdef LIBXML_OUTPUT_ENABLED
     printf("\tParse the XML files and output the result of the parsing\n");
+#else
+    printf("\tParse the XML files\n");
+#endif /* LIBXML_OUTPUT_ENABLED */
     printf("\t--version : display the version of the XML library used\n");
 #ifdef LIBXML_DEBUG_ENABLED
     printf("\t--debug : dump a debug tree of the in-memory document\n");
     printf("\t--shell : run a navigating shell\n");
     printf("\t--debugent : debug the entities defined in the document\n");
 #else
+#ifdef LIBXML_READER_ENABLED
     printf("\t--debug : dump the nodes content when using --stream\n");
+#endif /* LIBXML_READER_ENABLED */
 #endif
+#ifdef LIBXML_TREE_ENABLED
     printf("\t--copy : used to test the internal copy implementation\n");
+#endif /* LIBXML_TREE_ENABLED */
     printf("\t--recover : output what was parsable on broken XML documents\n");
     printf("\t--noent : substitute entity references by their value\n");
     printf("\t--noout : don't output the result tree\n");
     printf("\t--nonet : refuse to fetch DTDs or entities over network\n");
     printf("\t--htmlout : output results as HTML\n");
     printf("\t--nowrap : do not put HTML doc wrapper\n");
+#ifdef LIBXML_VALID_ENABLED
     printf("\t--valid : validate the document in addition to std well-formed check\n");
     printf("\t--postvalid : do a posteriori validation, i.e after parsing\n");
     printf("\t--dtdvalid URL : do a posteriori validation against a given DTD\n");
+    printf("\t--dtdvalidfpi FPI : same but name the DTD with a Public Identifier\n");
+#endif /* LIBXML_VALID_ENABLED */
     printf("\t--timing : print some timings\n");
     printf("\t--output file or -o file: save to a given file\n");
     printf("\t--repeat : repeat 100 times, for timing or profiling\n");
     printf("\t--insert : ad-hoc test for valid insertions\n");
+#ifdef LIBXML_OUTPUT_ENABLED
 #ifdef HAVE_ZLIB_H
     printf("\t--compress : turn on gzip compression of output\n");
 #endif
-#ifdef LIBXML_DOCB_ENABLED
-    printf("\t--sgml : use the DocBook SGML parser\n");
-#endif
+#endif /* LIBXML_OUTPUT_ENABLED */
 #ifdef LIBXML_HTML_ENABLED
     printf("\t--html : use the HTML parser\n");
 #endif
+#ifdef LIBXML_PUSH_ENABLED
     printf("\t--push : use the push mode of the parser\n");
+#endif /* LIBXML_PUSH_ENABLED */
 #ifdef HAVE_SYS_MMAN_H
     printf("\t--memory : parse from memory\n");
 #endif
     printf("\t--nowarning : do not emit warnings from parser/validator\n");
     printf("\t--noblanks : drop (ignorable?) blanks spaces\n");
+    printf("\t--nocdata : replace cdata section with text nodes\n");
+#ifdef LIBXML_OUTPUT_ENABLED
     printf("\t--format : reformat/reindent the input\n");
-    printf("\t--testIO : test user I/O support\n");
     printf("\t--encode encoding : output in the given encoding\n");
+    printf("\t--dropdtd : remove the DOCTYPE of the input docs\n");
+#endif /* LIBXML_OUTPUT_ENABLED */
+    printf("\t--nsclean : remove redundant namespace declarations\n");
+    printf("\t--testIO : test user I/O support\n");
 #ifdef LIBXML_CATALOG_ENABLED
     printf("\t--catalogs : use SGML catalogs from $SGML_CATALOG_FILES\n");
     printf("\t             otherwise XML Catalogs starting from \n");
@@ -1316,8 +1384,9 @@ static void usage(const char *name) {
 #endif
     printf("\t--loaddtd : fetch external DTD\n");
     printf("\t--dtdattr : loaddtd + populate the tree with inherited attributes \n");
-    printf("\t--dropdtd : remove the DOCTYPE of the input docs\n");
+#ifdef LIBXML_READER_ENABLED
     printf("\t--stream : use the streaming interface to process very large files\n");
+#endif /* LIBXML_READER_ENABLED */
     printf("\t--chkregister : verify the node registration code\n");
 #ifdef LIBXML_SCHEMAS_ENABLED
     printf("\t--relaxng schema : do RelaxNG validation against the schema\n");
@@ -1368,56 +1437,71 @@ main(int argc, char **argv) {
             noout = 1;
         } else 
 #endif
+#ifdef LIBXML_TREE_ENABLED
 	if ((!strcmp(argv[i], "-copy")) || (!strcmp(argv[i], "--copy")))
 	    copy++;
-	else if ((!strcmp(argv[i], "-recover")) ||
-	         (!strcmp(argv[i], "--recover")))
+	else
+#endif /* LIBXML_TREE_ENABLED */
+	if ((!strcmp(argv[i], "-recover")) ||
+	         (!strcmp(argv[i], "--recover"))) {
 	    recovery++;
-	else if ((!strcmp(argv[i], "-noent")) ||
-	         (!strcmp(argv[i], "--noent")))
+	    options |= XML_PARSE_RECOVER;
+	} else if ((!strcmp(argv[i], "-noent")) ||
+	         (!strcmp(argv[i], "--noent"))) {
 	    noent++;
-	else if ((!strcmp(argv[i], "-version")) ||
+	    options |= XML_PARSE_NOENT;
+	} else if ((!strcmp(argv[i], "-nsclean")) ||
+	         (!strcmp(argv[i], "--nsclean"))) {
+	    options |= XML_PARSE_NSCLEAN;
+	} else if ((!strcmp(argv[i], "-nocdata")) ||
+	         (!strcmp(argv[i], "--nocdata"))) {
+	    options |= XML_PARSE_NOCDATA;
+	} else if ((!strcmp(argv[i], "-nodict")) ||
+	         (!strcmp(argv[i], "--nodict"))) {
+	    options |= XML_PARSE_NODICT;
+	} else if ((!strcmp(argv[i], "-version")) ||
 	         (!strcmp(argv[i], "--version"))) {
 	    showVersion(argv[0]);
 	    version = 1;
 	} else if ((!strcmp(argv[i], "-noout")) ||
 	         (!strcmp(argv[i], "--noout")))
 	    noout++;
+#ifdef LIBXML_OUTPUT_ENABLED
 	else if ((!strcmp(argv[i], "-o")) ||
 	         (!strcmp(argv[i], "-output")) ||
 	         (!strcmp(argv[i], "--output"))) {
 	    i++;
 	    output = argv[i];
 	}
+#endif /* LIBXML_OUTPUT_ENABLED */
 	else if ((!strcmp(argv[i], "-htmlout")) ||
 	         (!strcmp(argv[i], "--htmlout")))
 	    htmlout++;
-#ifdef LIBXML_DOCB_ENABLED
-        else if ((!strcmp(argv[i], "-sgml")) ||
-		 (!strcmp(argv[i], "--sgml"))) {
-	    sgml++;
-	}
-#endif
+	else if ((!strcmp(argv[i], "-nowrap")) ||
+	         (!strcmp(argv[i], "--nowrap")))
+	    nowrap++;
 #ifdef LIBXML_HTML_ENABLED
 	else if ((!strcmp(argv[i], "-html")) ||
 	         (!strcmp(argv[i], "--html"))) {
 	    html++;
         }
 #endif /* LIBXML_HTML_ENABLED */
-	else if ((!strcmp(argv[i], "-nowrap")) ||
-	         (!strcmp(argv[i], "--nowrap")))
-	    nowrap++;
 	else if ((!strcmp(argv[i], "-loaddtd")) ||
-	         (!strcmp(argv[i], "--loaddtd")))
+	         (!strcmp(argv[i], "--loaddtd"))) {
 	    loaddtd++;
-	else if ((!strcmp(argv[i], "-dtdattr")) ||
+	    options |= XML_PARSE_DTDLOAD;
+	} else if ((!strcmp(argv[i], "-dtdattr")) ||
 	         (!strcmp(argv[i], "--dtdattr"))) {
 	    loaddtd++;
 	    dtdattrs++;
-	} else if ((!strcmp(argv[i], "-valid")) ||
-	         (!strcmp(argv[i], "--valid")))
+	    options |= XML_PARSE_DTDATTR;
+	}
+#ifdef LIBXML_VALID_ENABLED
+	else if ((!strcmp(argv[i], "-valid")) ||
+	         (!strcmp(argv[i], "--valid"))) {
 	    valid++;
-	else if ((!strcmp(argv[i], "-postvalid")) ||
+	    options |= XML_PARSE_DTDVALID;
+	} else if ((!strcmp(argv[i], "-postvalid")) ||
 	         (!strcmp(argv[i], "--postvalid"))) {
 	    postvalid++;
 	    loaddtd++;
@@ -1426,7 +1510,13 @@ main(int argc, char **argv) {
 	    i++;
 	    dtdvalid = argv[i];
 	    loaddtd++;
+	} else if ((!strcmp(argv[i], "-dtdvalidfpi")) ||
+	         (!strcmp(argv[i], "--dtdvalidfpi"))) {
+	    i++;
+	    dtdvalidfpi = argv[i];
+	    loaddtd++;
         }
+#endif /* LIBXML_VALID_ENABLED */
 	else if ((!strcmp(argv[i], "-dropdtd")) ||
 	         (!strcmp(argv[i], "--dropdtd")))
 	    dropdtd++;
@@ -1440,11 +1530,17 @@ main(int argc, char **argv) {
 	         (!strcmp(argv[i], "--auto")))
 	    generate++;
 	else if ((!strcmp(argv[i], "-repeat")) ||
-	         (!strcmp(argv[i], "--repeat")))
-	    repeat++;
+	         (!strcmp(argv[i], "--repeat"))) {
+	    if (repeat)
+	        repeat *= 10;
+	    else
+	        repeat = 100;
+	}
+#ifdef LIBXML_PUSH_ENABLED
 	else if ((!strcmp(argv[i], "-push")) ||
 	         (!strcmp(argv[i], "--push")))
 	    push++;
+#endif /* LIBXML_PUSH_ENABLED */
 #ifdef HAVE_SYS_MMAN_H
 	else if ((!strcmp(argv[i], "-memory")) ||
 	         (!strcmp(argv[i], "--memory")))
@@ -1455,9 +1551,12 @@ main(int argc, char **argv) {
 	    testIO++;
 #ifdef LIBXML_XINCLUDE_ENABLED
 	else if ((!strcmp(argv[i], "-xinclude")) ||
-	         (!strcmp(argv[i], "--xinclude")))
+	         (!strcmp(argv[i], "--xinclude"))) {
 	    xinclude++;
+	    /* options |= XML_PARSE_XINCLUDE; */
+	}
 #endif
+#ifdef LIBXML_OUTPUT_ENABLED
 #ifdef HAVE_ZLIB_H
 	else if ((!strcmp(argv[i], "-compress")) ||
 	         (!strcmp(argv[i], "--compress"))) {
@@ -1465,15 +1564,18 @@ main(int argc, char **argv) {
 	    xmlSetCompressMode(9);
         }
 #endif
+#endif /* LIBXML_OUTPUT_ENABLED */
 	else if ((!strcmp(argv[i], "-nowarning")) ||
 	         (!strcmp(argv[i], "--nowarning"))) {
 	    xmlGetWarningsDefaultValue = 0;
 	    xmlPedanticParserDefault(0);
+	    options |= XML_PARSE_NOWARNING;
         }
 	else if ((!strcmp(argv[i], "-pedantic")) ||
 	         (!strcmp(argv[i], "--pedantic"))) {
 	    xmlGetWarningsDefaultValue = 1;
 	    xmlPedanticParserDefault(1);
+	    options |= XML_PARSE_PEDANTIC;
         }
 #ifdef LIBXML_DEBUG_ENABLED
 	else if ((!strcmp(argv[i], "-debugent")) ||
@@ -1508,13 +1610,23 @@ main(int argc, char **argv) {
 	else if ((!strcmp(argv[i], "-format")) ||
 	         (!strcmp(argv[i], "--format"))) {
 	     noblanks++;
+#ifdef LIBXML_OUTPUT_ENABLED
 	     format++;
+#endif /* LIBXML_OUTPUT_ENABLED */
 	     xmlKeepBlanksDefault(0);
 	}
+#ifdef LIBXML_READER_ENABLED
 	else if ((!strcmp(argv[i], "-stream")) ||
 	         (!strcmp(argv[i], "--stream"))) {
 	     stream++;
 	}
+#endif /* LIBXML_READER_ENABLED */
+#ifdef LIBXML_SAX1_ENABLED
+	else if ((!strcmp(argv[i], "-sax1")) ||
+	         (!strcmp(argv[i], "--sax1"))) {
+	     sax1++;
+	}
+#endif /* LIBXML_SAX1_ENABLED */
 	else if ((!strcmp(argv[i], "-chkregister")) ||
 	         (!strcmp(argv[i], "--chkregister"))) {
 	     chkregister++;
@@ -1524,6 +1636,7 @@ main(int argc, char **argv) {
 	    i++;
 	    relaxng = argv[i];
 	    noent++;
+	    options |= XML_PARSE_NOENT;
 	} else if ((!strcmp(argv[i], "-schema")) ||
 	         (!strcmp(argv[i], "--schema"))) {
 	    i++;
@@ -1555,6 +1668,13 @@ main(int argc, char **argv) {
     }
 #endif
 
+#ifdef LIBXML_SAX1_ENABLED
+    if (sax1)
+        xmlSAXDefaultVersion(1);
+    else
+        xmlSAXDefaultVersion(2);
+#endif /* LIBXML_SAX1_ENABLED */
+
     if (chkregister) {
 	xmlRegisterNodeDefault(registerNode);
 	xmlDeregisterNodeDefault(deregisterNode);
@@ -1572,7 +1692,9 @@ main(int argc, char **argv) {
     if (dtdattrs)
 	xmlLoadExtDtdDefaultValue |= XML_COMPLETE_ATTRS;
     if (noent != 0) xmlSubstituteEntitiesDefault(1);
+#ifdef LIBXML_VALID_ENABLED
     if (valid != 0) xmlDoValidityCheckingDefaultValue = 1;
+#endif /* LIBXML_VALID_ENABLED */
     if ((htmlout) && (!nowrap)) {
 	xmlGenericError(xmlGenericErrorContext,
          "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\n");
@@ -1587,11 +1709,16 @@ main(int argc, char **argv) {
     }
 
 #ifdef LIBXML_SCHEMAS_ENABLED
-    if ((relaxng != NULL) && (stream == 0)) {
+    if ((relaxng != NULL)
+#ifdef LIBXML_READER_ENABLED
+        && (stream == 0)
+#endif /* LIBXML_READER_ENABLED */
+	) {
 	xmlRelaxNGParserCtxtPtr ctxt;
 
         /* forces loading the DTDs */
         xmlLoadExtDtdDefaultValue |= 1; 
+	options |= XML_PARSE_DTDLOAD;
 	if (timing) {
 	    startTimer();
 	}
@@ -1604,6 +1731,7 @@ main(int argc, char **argv) {
 	if (relaxngschemas == NULL) {
 	    xmlGenericError(xmlGenericErrorContext,
 		    "Relax-NG schema %s failed to compile\n", relaxng);
+            progresult = 5;
 	    relaxng = NULL;
 	}
 	xmlRelaxNGFreeParserCtxt(ctxt);
@@ -1625,6 +1753,7 @@ main(int argc, char **argv) {
 	if (wxschemas == NULL) {
 	    xmlGenericError(xmlGenericErrorContext,
 		    "WXS schema %s failed to compile\n", schema);
+            progresult = 5;
 	    schema = NULL;
 	}
 	xmlSchemaFreeParserCtxt(ctxt);
@@ -1644,11 +1773,18 @@ main(int argc, char **argv) {
             i++;
 	    continue;
         }
+#ifdef LIBXML_VALID_ENABLED
 	if ((!strcmp(argv[i], "-dtdvalid")) ||
 	         (!strcmp(argv[i], "--dtdvalid"))) {
 	    i++;
 	    continue;
         }
+	if ((!strcmp(argv[i], "-dtdvalidfpi")) ||
+	         (!strcmp(argv[i], "--dtdvalidfpi"))) {
+	    i++;
+	    continue;
+        }
+#endif /* LIBXML_VALID_ENABLED */
 	if ((!strcmp(argv[i], "-relaxng")) ||
 	         (!strcmp(argv[i], "--relaxng"))) {
 	    i++;
@@ -1664,25 +1800,39 @@ main(int argc, char **argv) {
 	/* Remember file names.  "-" means stdin.  <sven@zen.org> */
 	if ((argv[i][0] != '-') || (strcmp(argv[i], "-") == 0)) {
 	    if (repeat) {
-		for (acount = 0;acount < 100 * repeat;acount++)
-		    if (stream != 0)
+		xmlParserCtxtPtr ctxt = NULL;
+
+		for (acount = 0;acount < repeat;acount++) {
+#ifdef LIBXML_READER_ENABLED
+		    if (stream != 0) {
 			streamFile(argv[i]);
-		    else
-			parseAndPrintFile(argv[i]);
+		    } else {
+#endif /* LIBXML_READER_ENABLED */
+		        if (ctxt == NULL)
+			    ctxt = xmlNewParserCtxt();
+			parseAndPrintFile(argv[i], ctxt);
+#ifdef LIBXML_READER_ENABLED
+		    }
+#endif /* LIBXML_READER_ENABLED */
+		}
+		if (ctxt != NULL)
+		    xmlFreeParserCtxt(ctxt);
 	    } else {
+#ifdef LIBXML_READER_ENABLED
 		if (stream != 0)
 		    streamFile(argv[i]);
 		else
-		    parseAndPrintFile(argv[i]);
+#endif /* LIBXML_READER_ENABLED */
+		    parseAndPrintFile(argv[i], NULL);
 	    }
 	    files ++;
 	    if ((timing) && (repeat)) {
-		endTimer("100 iterations");
+		endTimer("%d iterations", repeat);
 	    }
 	}
     }
     if (generate) 
-	parseAndPrintFile(NULL);
+	parseAndPrintFile(NULL, NULL);
     if ((htmlout) && (!nowrap)) {
 	xmlGenericError(xmlGenericErrorContext, "</body></html>\n");
     }
